#include "sope.h"
#include "cdrom.h"
#include "hwr.h"
#include "assert.h"
#include "fileio.h"

#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/cdrom.h>
#include <linux/iso_fs.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#define btoi(b)		((b)/16*10 + (b)%16)
#define itob(i)		((i)/10*16 + (i)%10)

void incSector();
static int cd_strstr(const char *haystack,const char *needle,int n,int n1);
static int cd_faddr(const char *str,const char *file,int n);

char *coms[0x100] = {
	"Sync", "Nop", "Setloc", "Play", "Fordward", "Backward", "ReadN",
	"Standby",
	"Stop", "Pause", "Init", "Mute", "Demute", "Setfilter", "Setmode",
	"Getparam",
	"GetlocL", "GetlocP", "", "GetTN", "GetTD", "SeekL", "SeekP", "",
	"", "Test", "ID", "ReadS", "Reset", "", "ReadTOC", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", ""
};

void cd_misc()
{
	static int counth;
	static int vline;

	if (counth++ > 801) {
		counth = 0;
		vline++;
	}

	if (vline < 312) {
		HRM32(0x1110) = vline;
		return;
	}
	vline = 0;
	HRM32(0x1110) = 0;

/*	if (enable_vsync) {
	HRM32(i_reg) |= 1;
	cd_int = 1;
	}*/

	vsyncCount++;

	if (!(cdResultReady && cdResultC)) {
		if ((cdIrqNextTime >= 2) && (vsyncCount >= cdV)) {
			if (debug_cd) printf("irqnexttime>=2\n");
			cdResultP = 0;
			cdResultC = 1;
			cdResult[0] = 0x0;
			cdStat = 0;
			cdResultReady = 1;
			HRM32(i_reg) |= 0x4;
			cd_int=1;
			cdIrqNextTime = 0;
		}
		if ((cdIrqNextTime == 1) && (vsyncCount >= cdV)) {
//			printf("cdirqnexttime==1\n");
			cd_read(cdSetSector);
			incSector();
			cdResultP = 0;
			cdResultC = 1;
			cdResult[0] = 0x20;
			cdStat = 1;
			cdResultReady = 1;
			HRM32(i_reg) |= 0x4;
			cd_int=1;
			cdReaded = 0;
			if (ReadRetry) {
			cdV = vsyncCount + 1;
			}
			else {
			cdIrqNextTime = 0;
			}
		}
	}
}


void incSector()
{
	cdSetSector[2]++;
	if (cdSetSector[2] == 75) {
		cdSetSector[2] = 0;
		cdSetSector[1]++;
	}
	if (cdSetSector[1] == 60) {
		cdSetSector[1] = 0;
		cdSetSector[0]++;
	}
}

uint8 cdrom_read_reg0()
{
//	printf("read cd_reg0\n");
	if (cdResultReady && cdResultC) cdIoStatus |= 0x20;
	else cdIoStatus &= 0xdf;

	if (cdReaded == 0) return (cdIoStatus | 0x40);
	else return (cdIoStatus);
}

uint8 cdrom_read_reg1()
{
//	printf("read cd_reg1\n");
//	assert(cdResultReady);
	if (cdResultReady && cdResultC) HRM(cd_reg1) = cdResult[cdResultP];

	if (cdResultReady) {
		cdResultP++;
		if (cdResultP == cdResultC)
			cdResultReady = 0;
	}
	return (HRM(cd_reg1));
}

uint8 cdrom_read_reg2()
{
//	assert(0);
	return 0;
}

uint8 cdrom_read_reg3()
{
//	printf("read cd_reg3\n");
	return (cdStat);
}

void cdrom_write_reg0(unsigned char cdc)
{
//	printf("written 0x%2.2x to cd_reg0\n",cdc);
	if (cdc == 0) {
		cdParamP = 0;
		cdResultReady = 0;
		return;
	}
	if (cdc & 0x01)
		cdResultReady = 1;
	if (cdc == 0x01)
		cdReset = 1;
}

#define COMMON_END	\
cdResultP = 0;	\
cdResultC = 1;	\
cdStat = 2;	\
cdResult[0] = 2;	\
cdResultReady = 1;

void cdrom_write_reg1(unsigned char cdc)
{
	int i;

	if (debug_cd) printf("cd command: \"%s\",0x%x,pc0=%x\n", coms[cdc], cdc, ctx.pc0);

	switch (cdc) {
	case CdlSetloc:
		for (i = 0; i < 3; i++)
			cdSetSector[i] = btoi(cdParam[i]);
		cdSetSector[3] = 0;
		COMMON_END;
		break;
	case CdlPause:
		if (cdIrqNextTime == 1) {
			cdV = vsyncCount + 1;
			cdIrqNextTime = 2;
		}
		COMMON_END;
		break;
	case CdlInit:
		cdMode = 0x00;
		COMMON_END;
		cd_init();
//		if (cd_init()); {cdStat = 5; cdResult[0] = 1;}
		break;
	case CdlStop:
		cd_stop();
	case CdlSync:
	case CdlNop:
	case CdlDemute:
	case CdlSeekP:
	case CdlSeekL:
		COMMON_END;
		break;
	case CdlSetmode:
		if (cdMode != cdParam[0]) {
		cdMode = cdParam[0];
		if (debug_cd) printf("setting cdrom to mode == %2.2x\n", cdMode);
		cd_setmode(cdMode);
		}
		COMMON_END;
		break;
	case CdlReadS:
		ReadRetry = 0;
	case CdlReadN:
		if (cdc == CdlReadN) ReadRetry = 1;
		cdIrqNextTime = 1;
		cdV = vsyncCount + 1;
		cdReaded = 1;
		COMMON_END;
		break;
/*	case CdlGetlocP:
		cdResultP = 0;
		cdResultC = 8;
		cdResult[0] = itob(0); // track
		cdResult[1] = itob(0); // index
		cdResult[2] = itob(cdSetSector[0]); // min
		cdResult[3] = itob(cdSetSector[1]); // sec
		cdResult[4] = itob(cdSetSector[2]); // frame
		cdResult[5] = itob(0); // amin
		cdResult[6] = itob(0); // asec
		cdResult[7] = itob(0); // aframe
		cdStat = 2;
		cdResultReady = 1;
		break;*/
	case CdlGetTD:
		cdTrack = btoi(cdParam[0]);
		cd_getTD(cdTrack + 1, cdResultTD);
		cdResultP = 0;
		cdResultC = 4;
		cdStat = 2;
		cdResult[0] = 2;
		cdResult[1] = itob(cdResultTD[1]);
		cdResult[2] = itob(cdResultTD[2]);
		cdResult[3] = itob(cdResultTD[3]);
		cdResultReady = 1;
		break;
	case CdlGetTN:
		cd_getTN(cdResultTN);
		cdResultP = 0;
		cdResultC = 3;
		cdStat = 2;
		cdResult[0] = 2;
		cdResult[1] = itob(cdResultTN[0]);
		cdResult[2] = itob(cdResultTN[1]);
		cdResultReady = 1;
		break;
	case CdlPlay:
		cd_stop();
		cd_play(cdSetSector);
		cdResultP = 0;
		cdResultC = 1;
		cdStat = 2;
		cdResult[0] = 2;
		cdResultReady = 1;
		break;
	default:
		printf("cdrom command %x(%d) not implemented\n", cdc, cdc);
		return;
	}

	cd_int = 1;
	ctx.nstp = 0x10000;
	HRM32(i_reg) |= 0x4;
}

void cdrom_write_reg2(uint8 cdc)
{
//	printf("written 0x%2.2x to cd_reg2\n",cdc);
	if ((cdReset == 2) && (cdc == 7)) {
		/*This means that 1 in r0 and 7 in r3 have been written*/
		cdParamP = 0;
		cdResultReady = 1;
		cdReset = 0;
		cdStat = 0;
		cdIoStatus = 0;
		if (cdResultReady && cdResultC) cdIoStatus |= 0x20;
		else cdIoStatus &= 0xdf;

		return;
	}
	/*Used for sending argument*/
	cdReset = 0;
	cdParam[cdParamP++] = cdc;
	if (cdParamP > 7) {
		cdParamP = 7;
	}
}

void cdrom_write_reg3(uint8 cdc)
{
//	printf("written 0x%2.2x to cd_reg3\n",cdc);
	if ((cdc == 7) && (cdReset == 1))
		/*This means there is a chance that a reset might be in preparation*/
		cdReset = 2;
	else
		cdReset = 0;

/*	if (cdResultReady && (cdc  == 7)) {
		cdResultC = 0;
		cdStat = 0;
	}*/
}


void cd_reset() {
	cdReset = 0;
	cd_int = 0;
	cdIrqNextTime = 0;
	cdReaded = -1;
	lastread = 0;
	memset(cd_buffer, 0, sizeof(cd_buffer));
}


void cddma() {
	uint32 addr;
	uint32 size;
	int i;

	addr = HRM32(cd_madr);
	size = HRM32(cd_bcr); // bit from 31 - 16 = nblocks , 15 - 0 = sizeblock
	size = ((size & 0xffff) * 4) * (size >> 16); 
	addr &= 0x1fffff;

	if (size == 12)
	    for (i = 0; i < size; i++)
		RAOM[addr++] = cd_buffer[i+(lastread*DATA_SIZE)];
	else
	    for (i = 0; i < size; i++)
		RAOM[addr++] = cd_buffer[i+(lastread*DATA_SIZE+12)]; //skip header and subheader

	HRM32(cd_chcr) &= 0xfeffffff;
	HRM32(i_reg) |= 8;  // DMA finished
	HRM32(0x10f4) = 0x08000000;

	if (debug_cd) printf("cd dma: transferred %u bytes, to address %8.8x\n"
	,size,HRM32(cd_madr));
}
int cdHandle=-1;

void cd_read(char *param)  {
	union {
	    struct cdrom_msf msf;
	    uint8 buf[CD_FRAMESIZE_RAW];
	} cr;
	int i;
	char param0[3];

	if (cdHandle<1) return;

	if (debug_cd) printf ("reading %x:%x:%x\n",param[0],param[1],param[2]);

	for (i=0;i<3;i++) param0[i] = itob(param[i]);

	for (i=0;i<BUFFER_SIZE;i++)
	    {
	    if (!memcmp(cd_buffer+(i*DATA_SIZE),param0,3)) {
//		printf ("%d\n",i);
//		printf ("%x:%x:%x\n",cd_buffer[i*DATA_SIZE],cd_buffer[i*DATA_SIZE+1],cd_buffer[i*DATA_SIZE+2]);
		lastread = i;
		return;
		}
	    }

	cr.msf.cdmsf_min0=param[0];
	cr.msf.cdmsf_sec0=param[1];
	cr.msf.cdmsf_frame0=param[2];
	if (ioctl(cdHandle, CDROMREADRAW, &cr)) {
    	    printf("IOCTL ERROR #%i,%s\n",errno,strerror(errno));
//    	    ciao("");
	    memset (cd_buffer+(lastread*DATA_SIZE),0,CD_FRAMESIZE_RAW);
	    return;
	}
	if (++lastread >= BUFFER_SIZE) lastread = 0;
	memcpy(cd_buffer+(lastread*DATA_SIZE),cr.buf+12,CD_FRAMESIZE_RAW-12); // skip sync data
}


void cd_setmode(char mode) {
	switch (cdMode & 0x30) {
	    case 0x00: // 00 - 2048
			if (debug_cd) printf ("sector size == 2048\n"); break;
	    case 0x10: // 01 - 2328
			if (debug_cd) printf ("sector size == 2328\n"); break;
	    case 0x20: // 10 - 2340
	    case 0x30: // 11 - 2340
			if (debug_cd) printf ("sector size == 2340\n"); break;
	    }
}

void cd_getTD(int nr,char *result)
{
struct cdrom_tocentry entry;

if (cdHandle<1) return;

entry.cdte_track = nr;
entry.cdte_format = CDROM_MSF;

if (ioctl(cdHandle,CDROMREADTOCENTRY,&entry)==-1) return;

result[1] = entry.cdte_addr.msf.minute;
result[2] = entry.cdte_addr.msf.second;
result[3] = entry.cdte_addr.msf.frame;
}


void cd_getTN(char *result)
{
struct cdrom_tochdr toc;

if (cdHandle<1) return;

if (ioctl(cdHandle,CDROMREADTOCHDR,&toc)==-1) return;

result[0] = toc.cdth_trk0;
result[1] = toc.cdth_trk1;
}


int cd_init() {
	if (cdHandle > 0) return 0; // it's already open
	cdHandle = open("/dev/cdrom", O_RDONLY | O_NONBLOCK);
	if (cdHandle == -1) {printf ("could not open cdrom\n"); return 1;}

	if (ioctl(cdHandle,CDROMSTART,NULL)) {
	    printf("IOCTL ERROR #%i\n",errno);
	    ciao("");
	}
	cd_stop();
	return 0;
}


void cd_close() {
	cd_stop();
	if (cdHandle!=-1) close(cdHandle);
	cdHandle = -1;
}


void cd_play(char *param) {
	struct cdrom_msf addr;
	char ptmp[4];

	if (cdHandle<1) return;

	cd_getTN(ptmp);
	cd_getTD(ptmp[1],ptmp);

	addr.cdmsf_min0 = param[0];
	addr.cdmsf_sec0 = param[1];
	addr.cdmsf_frame0 = param[2];
	addr.cdmsf_min1 = ptmp[1];
	addr.cdmsf_sec1 = ptmp[2];
	addr.cdmsf_frame1 = ptmp[3];

	if (ioctl(cdHandle,CDROMPLAYMSF,&addr)==-1)
	    printf ("# Sope # error playing cd-da: min=%d,sec=%d,frame=%d\n",
	    addr.cdmsf_min0,addr.cdmsf_sec0,addr.cdmsf_frame0);
}


void cd_stop()
{
struct cdrom_subchnl sc;

if (cdHandle<1) return;

sc.cdsc_format = CDROM_MSF;
if (ioctl(cdHandle,CDROMSUBCHNL,&sc)==-1) return;

switch (sc.cdsc_audiostatus)
    {
    case CDROM_AUDIO_PAUSED:
    case CDROM_AUDIO_PLAY: ioctl(cdHandle,CDROMSTOP); break;
    }
}


void cd_setvol(){}


void cd_boot() {
	struct iso_primary_descriptor *isods;
	int faddr,laddr;
	char temp[0x800],file[12];

	cd_init();
	if (cdHandle==-1) ciao("");

	lseek(cdHandle,0x8000,SEEK_SET);
	read(cdHandle,temp,0x800); // reads cdrom super block
	isods = (struct iso_primary_descriptor *) temp;

	if ((strncmp(isods->id,"CD001",5)) |
	    (strncmp(isods->application_data+141,"CD-XA001",8))/* |
	    (strncmp(isods->system_id,"PLAYSTATION",11))*/)
	    ciao("are you sure this is a psx cd? (or maybe i'm wrong)");

	printf ("# Sope # Running CD \"%s\"\n",isods->volume_id);

	laddr = lseek(cdHandle,0x800*5,SEEK_CUR);
	read(cdHandle,temp,0x800);

	faddr = cd_faddr(temp,"SYSTEM.CNF",10);
	printf ("SYSTEM.CNF ok\n");

	lseek(cdHandle,faddr,SEEK_SET);
	read(cdHandle,temp,0x800);
	{
	int i=6;
	char *ttmp;

	if ((ttmp = strstr(temp,"cdrom:"))==NULL)
	    ciao("error running cd (cdrom:)");

	if (ttmp[6]=='\\') i++;

	lseek(cdHandle,faddr+(ttmp-temp)+i,SEEK_SET);
	}

	read(cdHandle,file,11);
	lseek(cdHandle,laddr,SEEK_SET);
	read(cdHandle,temp,0x800);

	faddr = cd_faddr(temp,file,11);
	lseek(cdHandle,faddr,SEEK_SET);

	{
	EXE_HEADER header;

	read (cdHandle,&header,sizeof (header));
	printf ("%s %x %x\n",header.id,header.t_addr,ctx.pc0);
	if (strncmp(header.id,"PS-X EXE",8)) ciao("error on file");

	lseek(cdHandle,2048-sizeof(header),SEEK_CUR);

	read (cdHandle,&RAOM[header.t_addr&0x1fffff],header.t_size); 
	ctx.pc0 = header.pc0;
	ctx.gpr[28] = header.GP0;
	ctx.gpr[29] = header.s_size;
	}

	cd_close();
}


static int cd_strstr(const char *haystack,const char *needle,int n,int n1)
{
int i;

for (i=0;i<n;i++)
    {
    if (!strncmp(haystack+i,needle,n1)) return i;
    }
return -1;
}


static int cd_faddr(const char *str,const char *file,int n)
{
struct iso_directory_record *isodr;
int addr;


addr = cd_strstr(str,file,0x800,n);
if (addr==-1)
    ciao("error running cd (-1)");

addr -= sizeof(struct iso_directory_record);
isodr = (struct iso_directory_record *) (str+addr);

return (((int *)isodr->extent)[0]<<11);
}

