#include "sope.h"
#include "gpu.h"
#include "gpu_defs.h"
#include "renderer.h"
#include "memio.h"
#include "vram.h"

#include "gui.h"
/*
  gp0mode:
  0:	get arguments for primitive
  1:	will get a primitive
  2:	cpu2gpu transfert
*/

void moveimage();

int unsigned cp,cpnum;
int unsigned c2g_w,c2g_h,c2g_x,c2g_y,c2g_c;

void primitive(int unsigned af);
void execprim(char unsigned pp);

/*GPU primitives*/
#include "gpu_getvars.h"
#include "gpu_sprites.h"
#include "gpu_tiles.h"
#include "gpu_polys.h"
#include "gpu_lines.h"
#include "gpu_misc.h"

uint32 gp0read() {
	int unsigned cc;
	switch(gp0mode) {
	case 2:	
		GPU_DEBUG(("GP0 reading transfert data\n"));
		cc=vram32(c2g_x/2+512*c2g_y+(c2g_c%(c2g_w/2))+(c2g_c/(c2g_w/2))*512);
		c2g_c+=2;
		if (c2g_c==(c2g_w*c2g_h)) {
			HRM32(0x1814)&=0xf7ffffff;
			gp0mode=0;
		}
		return(cc);
		break;
	default:
		return(HRM32(0x1810));
	}
}

void gp0write(uint32 af) {
  
	switch(gp0mode) {
	case 0:	
		GPU_DEBUG(("GP0 pak data %8.8x\n",af));
		pak[cpnum]=af;if (cpnum==cpmax[cp]) execprim(cp);cpnum++;break;
	case 1: 
		GPU_DEBUG(("GP0 primitive %8.8x\n",af));
		primitive(af);break;
	case 2:	
		GPU_DEBUG(("GP0 transfert data %8.8x\n",af));
		glbuf[c2g_x+(c2g_c%c2g_w)+1024*(c2g_y+(c2g_c/c2g_w))]=hitt(af&0xffff);
		c2g_c++;
		if (c2g_c==(c2g_w*c2g_h)) {
			HRM32(0x1814)&=0xf7ffffff;
			gp0mode=0;break;
		}
		glbuf[c2g_x+(c2g_c%c2g_w)+1024*(c2g_y+(c2g_c/c2g_w))]=hitt(af>>16);
		c2g_c++;
		if (c2g_c==(c2g_w*c2g_h)) {
			HRM32(0x1814)&=0xf7ffffff;
			gp0mode=0;
		}
		break;
	}
}

void gpufunc(uint32 af) {
	int unsigned afu;

	GPU_DEBUG(("GP1 command %8.8x\n",af));
	afu=af>>24;
	switch (afu) {
	case 0x00:  break;/*All Reset*/
	case 0x01:  break;/*Comm Reset*/
	case 0x02:	break;/*IRQ Reset*/
	case 0x03:	break;
	case 0x04:  
		if ((af&0xffffff)==0) gp0mode=1;
		if ((af&0xffffff)==2) gp0mode=1;
		break;/*DMA Setup*/
	case 0x05:  
		gs.dofs_x=af&0x3ff;
		gs.dofs_y=(af&0x7fc00)>>10;
		GPU_DEBUG(("Display offsets: %u %u\n",gs.dofs_x,gs.dofs_y));
		break;/*Start of Display Area*/
	case 0x06:	break;
	case 0x07:	break;
	case 0x08:	initscreen(af&0xff);break;
	case 0x10:	
		switch(af&0xff) {
			case 3:	HRM32(0x1810)=gs.clip_x+(gs.clip_y<<10);break;
			case 4:	HRM32(0x1810)=gs.clip_w+(gs.clip_h<<10);break;
			case 5:	HRM32(0x1810)=gs.ofs_x+(gs.ofs_y<<11);break;
			case 7:	HRM32(0x1810)=2;break;
			default:printf("# Sope # Info? %x\n",af&0xff);break;
			} 
				break;
	default:	GPU_DEBUG(("GPU command %8.8x not implemented\n",af));
				break;
	}
}

void execprim(uint8 pp)
{
	int i;
	if ((pp>=0x20) && (pp<0x80)) dsp_changed=1;
	if (pp==0xa0) dsp_changed=1;
	gp0mode=1;
	/*Drawing primitives*/
	if (norender==0)
		switch (pp&0xfc){
		case 0x20:	draw_fs_triangle((pp>>1)&1);break;
		case 0x24:	draw_fst_triangle((pp>>1)&1,pp&1);break;
		case 0x28:	draw_fs_quadrangle((pp>>1)&1);break;
		case 0x2c:	draw_fst_quadrangle((pp>>1)&1,pp&1);break;
		case 0x30:  draw_gs_triangle((pp>>1)&1);break;
		case 0x34:  draw_gst_triangle((pp>>1)&1,pp&1);break;
		case 0x38:  draw_gs_quadrangle((pp>>1)&1);break;
		case 0x3c:  draw_gst_quadrangle((pp>>1)&1,pp&1);break;
		case 0x40:	draw_fs_line((pp>>1)&1);break;
		case 0x50:  draw_gs_line((pp>>1)&1);break;
		case 0x60:	draw_tile(0,0,(pp>>1)&1);break;
		case 0x64:	draw_sprite(0,0,(pp>>1)&1,pp&1);break;
		case 0x68:	draw_tile(1,1,(pp>>1)&1);break;
		case 0x70:	draw_tile(8,8,(pp>>1)&1);break;
		case 0x74:  draw_sprite(8,8,(pp>>1)&1,pp&1);break;
		case 0x78:  draw_tile(16,16,(pp>>1)&1);break;
		case 0x7c:  draw_sprite(16,16,(pp>>1)&1,pp&1);break;
		}
	/*General primitives*/
	switch (pp){
	case 0x00:	GPU_DEBUG(("Nothing...\n");break);
	case 0x01:	rnd->clear_texture_cache();break;
	case 0x02:	draw_blockfill();break;
	case 0x80: 	moveimage();break;
	case 0xa0:	
		HRM32(0x1814)|=0x08000000;
		gp0mode=2;c2g_x=pak[1]&0xffff;c2g_y=pak[1]>>16;
		c2g_c=0;c2g_w=(pak[2]&0xffff);c2g_h=pak[2]>>16;
/*		if ( 
		( (c2g_x>=_texx) && (c2g_x<(_texx+256)) )
		||	( ((c2g_x+c2g_w)>=_texx) && ((c2g_x+c2g_w)<(_texx+256)) )
		||	( ((c2g_x+c2g_w)>=(_texx+256)) && (c2g_x<_texx) )
		) {
		if ( 
		( (c2g_y>=_texy) && (c2g_y<(_texy+256)) )
		||  ( ((c2g_y+c2g_h)>=_texy) && ((c2g_y+c2g_h)<(_texy+256)) )
		||  ( ((c2g_y+c2g_h)>=(_texy+256)) && (c2g_y<_texy) )
		) { 
		_texx=0xffffffff;_texy=0xffffffff;
		}
		}*/
		if (debug_dma) 
			printf("CPU 2 GPU transfert %u,%u  %u,%u\n",c2g_x,c2g_y,c2g_w,c2g_h);
		break;
	case 0xc0:	
		HRM32(0x1814)|=0x08000000;
		gp0mode=2;c2g_x=pak[1]&0xffff;c2g_y=pak[1]>>16;
		c2g_c=0;c2g_w=(pak[2]&0xffff);c2g_h=pak[2]>>16;
		if (debug_dma) printf("GPU 2 CPU transfert %u,%u  %u,%u\n",c2g_x,c2g_y,c2g_w,c2g_h);
		break;
	case 0xe1:	
		i=pak[0]&0xffff;
		gs.tp_y=((i>>4)&1)*256;
		gs.tp_x=64*(i&0xf);
		gs.abr=(i>>5)&3;rnd->set_blending();
		i=(i>>7)&3;
		if (i<3)
			gs.tp_pal=i;
		else
		{
			gui->print("Invalid tp_pal");gs.tp_pal=2;
		}
		gs.dtd=(i>>9)&1;
		gs.dfe=(i>>10)&1;
/*		md=(i>>11)&1;_md_=md*0xFF000000;
		me=(i>>12)&1;_me_=me*0xFF000000;*/
		break;
	case 0xe3:	
		gs.clip_x=pak[0]&0x3ff;gs.clip_y=(pak[0]&0x000ffc00)>>10;
		if (gs.clip_x<0) gs.clip_x=0;
		if (gs.clip_y<0) gs.clip_y=0;
		gs.oxi=gs.clip_x;gs.oyi=gs.clip_y;
		rnd->set_scissor();
		break;
	case 0xe4:	
		gs.clip_w=pak[0]&0x3ff;gs.clip_h=(pak[0]&0x000ffc00)>>10;
		if (gs.clip_w>1023) gs.clip_w=1023;
		if (gs.clip_h>511) gs.clip_h=511;
		gs.oxf=gs.clip_w+1;gs.oyf=gs.clip_h+1;
		rnd->set_scissor();
		break;
	case 0xe5:	gs.ofs_x=pak[0]&0x7ff;gs.ofs_y=(pak[0]&0x003ff800)>>11;
		GPU_DEBUG(("Drawing offsets: +%u+%u\n",gs.ofs_x,gs.ofs_y));
		break;
	case 0xe2:	
		gs.tww=256-((pak[0]<<3)&0xf8);
		gs.twh=256-((pak[0]>>2)&0xf8);
		gs.twx=((pak[0]>>07)&0xf8);
		gs.twy=((pak[0]>>12)&0xf8);
		GPU_DEBUG(("Texture window settings: %ux%u+%u+%u\n",gs.tww,gs.twh,gs.twx,gs.twy));
		break;				
	case 0xe6:	
		gs.mask1=pak[0]&1;gs._md_=0xFF000000*gs.mask1;
		gs.mask2=(pak[0]>>1)&1;gs._me_=0xFF000000*gs.mask2;
		GPU_DEBUG(("Mask settings %u %u\n",gs.mask1,gs.mask2));
		break;				
	default	 :	GPU_DEBUG(("Primitive %2.2x unknown...\n",pp));
		break;
	}
}

void primitive(int unsigned af) {
	cp=af>>24;
	if (cpmax[cp]==9) {
		printf("# Sope # ---%8.8x--- GPU primitive %2.2x not implemented\n",af,cp);
//		ciao("");
	} else {
		gp0mode=0;
		pak[0]=af;
		if (cpmax[cp]==0) execprim(cp);
		cpnum=1;
	}
}

void gpudma() {
	int unsigned i,j,k,l;
	uint16 *su;
	uint32 dma1,dma2,dma3;
#ifdef GRAND_ENDIAN
	uint16 buf;
#endif
	dma1=HRM32(0x10a0);
	dma2=HRM32(0x10a4);
	dma3=HRM32(0x10a8);
	switch(dma3) {
	case 0x01000200:if (debug_dma) printf("VRAM2MEM %8.8x,%8.8x\n",dma1,dma2);
		su=RAOM16+((dma1&0x1fffff)/2);
		for (i=0;i<((dma2>>16)*32);i++) {
#ifdef GRAND_ENDIAN
			buf=tthi(glbuf[c2g_x+(c2g_c%c2g_w)+1024*(c2g_y+(c2g_c/c2g_w))]);
			*su=END_16((buf));
#else
			*su=tthi(glbuf[c2g_x+(c2g_c%c2g_w)+1024*(c2g_y+(c2g_c/c2g_w))]);
#endif
			su++;c2g_c++;
		}
		gp0mode=0;
		HRM32(0x1814)&=0xf7ffffff;
		break;
	case 0x01000201:if (debug_dma) printf("MEM2VRAM %8.8x,%8.8x\n",dma1,dma2);
		su=RAOM16+((dma1&0x1fffff)/2);
		for (i=0;i<((dma2>>16)*32);i++) {
#ifdef GRAND_ENDIAN
			buf=END_16((*su));
			glbuf[c2g_x+(c2g_c%c2g_w)+1024*(c2g_y+(c2g_c/c2g_w))]=hitt(buf);
#else
			glbuf[c2g_x+(c2g_c%c2g_w)+1024*(c2g_y+(c2g_c/c2g_w))]=hitt(*su);
#endif
			su++;c2g_c++;
		}
		gp0mode=0;
		HRM32(0x1814)&=0xf7ffffff;
		break;
	case 0x01000401:l=dma1;
		do {
			k=rm_32(l);
			j=k>>24;
			gp0mode=1;
			for (i=0;i<j;i++) gp0write(rm_32(l+4*(i+1)));
			l=k&0xffffff;
		} while ((l!=0xffffff)&&(l!=0x000000));
		break;
	default:if (debug_dma) printf("GPU DMA function %8.8x not impl...\n",dma3);
	}
	HRM32(0x10a8)&=0xfeffffff;
	if (debug_dma) printf("DMA transfer OK\n");
}


void otcdma() {
	int unsigned i;
	uint32 dma1,dma2,dma3;
	dma1=HRM32(0x10e0);
	dma2=HRM32(0x10e4);
	dma3=HRM32(0x10e8);

	switch(dma3) {
	case 0x11000002:	
		if (debug_dma) printf("Clear Reverse stuff %u\n",dma2);
		for (i=0;i<dma2;i++)
#ifdef GRAND_ENDIAN
			wm_32((dma1&0x1fffff)-4*(i+1),dma1-4*i);
#else
			getopc(dma1-i*4)=(dma1&0x1fffff)-(i+1)*4;
#endif
		//wm_32((dma1&0x1fffff)-4*(i+1),dma1-4*i);
		break;
	case 0x00000000:	break;
	default:		printf("OTC DMA function %8.8x not impl...\n",dma3);
	}
	HRM32(0x10e8)&=0xfeffffff;
}

void moveimage(){
	short signed xs,ys,xd,yd,w,h;
	int i,j,txi,txf,tyi,tyf;
	xs=sspak(2);
	ys=sspak(3);
	xd=sspak(4);
	yd=sspak(5);
	w=sspak(6);
	h=sspak(7);

	GPU_DEBUG(("Moveimage %u,%u ---> %u,%u : (%u,%u)\n",xs,ys,xd,yd,w,h));

	if ((yd<(1-h)) || (xd<(1-w)) || (yd>511) || (xd>1023)) return;
	if ((ys<(1-h)) || (xs<(1-w)) || (ys>511) || (xs>1023)) return;
	txi=0;txf=w;tyi=0;tyf=h;
	if (yd<0) tyi=-yd;
	if (yd>(512-h)) tyf=h-(yd-(512-h));
	if (xd<0) txi=-xd;
	if (xd>(1024-w)) txf=w-(xd-(1024-w));

	for (j=tyi;j<tyf;j++) {
		for (i=txi;i<txf;i++) {
			glbuf[1024*(yd+j)+xd+i]=glbuf[1024*(ys+j)+xs+i];}
	}	 
}


void screenshot(int s) {
	FILE *f;
	int i=0,x=0,w,h;
	char file[20],header[32];

	while (i<1000) {
	    sprintf (file,"screen%3.3d.ppm",i);
	    if ((f=fopen(file,"r"))==NULL) {
		if ((f=fopen(file,"w"))==NULL) {
		    printf ("# Sope # Error creating %s\n",file);
		    return;
		}
		i=1000;
	    }
	    i++;
	}

	printf ("# Sope # Creating screenshot %s\n",file);

	if (s) {
	    w = width;
	    h = height;
	}
	else {
	    w = 1024;
	    h = 512;
	}

	sprintf (header,"P6\n%d %d\n255\n",w,h);
	fputs(header,f);

	for (i=0;i<h;i++)
	    for (x=0;x<w;x++) {
		fputc((glbuf[(i<<10)+x])&0xff,f);
		fputc((glbuf[(i<<10)+x]>>8)&0xff,f);
		fputc((glbuf[(i<<10)+x]>>16)&0xff,f);
	    }

	fclose(f);
}
