#include "sope.h"
#include "hwr.h"
#include "cdrom.h"
#include "spu.h"
#include "gpu.h"
#include "mdec.h"

void otcdma();

void set_root_counter(uint16 hrn,uint32 af) {
	switch (hrn%16){
		case 0:
//			printf("Setting counter %u value to %d\n",(hrn>>4)&3,af);
			break;
		case 4:
//			printf("Setting counter %u mode to %x\n",(hrn>>4)&3,af);
			break;
		case 8:
//			printf("Setting counter %u target to %d\n",(hrn>>4)&3,af);
			break;
	}

}

uint32 hread(uint16 hrn) {
	if ((debug_hre)/* && (hrn<0x1070)*/) printf("Reading HR %4.4x,%8.8x\n",hrn,HRM32(hrn));

	switch(hrn) {
	case gp0:	return(gp0read());
	case t1_count:	return((ctx.nstp*480)/ncyc);
	case mdec_ctrl: printf ("mdec(read): on mdec_ctrl\n"); break;
	case mdec_stat: return mdec_readstat();
	}
	return(HRM32(hrn));
}

uint16 hread_word(uint16 hrn) {
	if (debug_hre) printf("Reading word HR %4.4x\n",hrn);

	switch(hrn) {
	case 0x1044: return 0xffff;
/*	case sp0:	printf("reading sp0 <-- %4.4x\n",HRM16(hrn));break;
	case sp1:	printf("reading sp1 <-- %4.4x\n",HRM16(hrn));break;
	case sp2:   printf("reading sp2 <-- %4.4x\n",HRM16(hrn));break;*/
	}
	return(HRM16(hrn));
}

void hwrite(uint32 af,uint16 hrn) {
	if (debug_hwr) printf("putting %8.8x in HR %4.4x\n",af,hrn);
	switch(hrn) {
		case i_reg:	HRM32(i_reg)&=af; return;
//		case i_mask:HRM32(hrn)|=af;return;
		case gp1:	gpufunc(af);return;
		case gp0:	gp0write(af);break;
	}
	HRM32(hrn)=af;
	switch(hrn) {
//		case dv5_delay:	printf("dv5_delay %u\n",af);break;
//		case com_delay: printf("com_delay %u\n",af);break;
/*DMA stuff*/
//		case cd_madr:	printf("madr %8.8x\n",af);break;
//		case cd_bcr:	printf("bcd %8.8x\n",af);break;
		case mdec_ctrl: printf ("mdec:af=%8.8x on mdec_ctrl\n",af); break;
		case mdec_stat: mdec_writestat(af); break;
		case mdecin_chcr:
			if (HRM32(d_pcr)&0x8) {
				if (debug_dma) printf ("CPU -> MDEC DMA Transfer\n");
				mdecindma();
			}
			return;
		case mdecout_chcr:
			if (HRM32(d_pcr)&0x80) {
				if (debug_dma) printf ("MDEC -> CPU DMA Transfer\n");
				mdecoutdma();
			}
			return;
		case gpu_chcr:
			if (HRM32(d_pcr)&0x800) {
				if (debug_dma) printf("GPU DMA Transfer\n");
				gpudma();
			}
			return;
		case cd_chcr:
//			printf("chcr %8.8x\n",af);
			if (HRM32(d_pcr)&0x8000) {
				if (debug_dma) printf("CD DMA Transfer\n");
				cddma();
			}
			return;
		case spu_chcr:	
			if (HRM32(d_pcr)&0x80000) {
				if (debug_dma) printf("SPU DMA Transfer\n");
				spudma();
			}
			return;
		case pio_chcr:
			if (HRM32(d_pcr)&0x800000) {
				if (debug_dma) printf("PIO DMA Transfer\n");
			}
			return;
		case otc_chcr:	
			if (HRM32(d_pcr)&0x8000000) {
				if (debug_dma) printf("OTC DMA Transfer\n");
				otcdma();
			}
			return;
/*Root counters*/
		case t0_count: 
		case t0_mode:
		case t0_target:
		case t1_count: 
		case t1_mode:
		case t1_target:
		case t2_count: 
		case t2_mode:
		case t2_target:
			set_root_counter(hrn,af);
			break;
		case d_icr: 
//			if (af & 0x08000000) cdIrqNextTime=1;
			HRM32(d_icr)&=((~af)&0xff000000);
			return;

	}
}

void hwrite_word(uint16 af,uint16 hrn) {
	if ((hrn>=0x1c00) && (hrn<0x1d80)) setvoiceparam(hrn,af);
	else	{if (debug_hwr) printf("putting %4.4x in HR %4.4x\n",af,hrn);}
	
	switch(hrn) {
		case i_reg: HRM16(i_reg)&=af;return;
//		case i_mask:  HRM16(hrn)|=af;return;
/*SPU stuff*/
		case spu_keyON_1:vonof(af,1);break;
		case spu_keyON_2:vonof(af,2);break;
		case spu_keyOFF_1:vonof(af,3);break;
		case spu_keyOFF_2:vonof(af,4);break;
		case spu_sb_addr:
			spuptr=(uint16 *)(SBF+(af*8));
			break;
		case spu_data:
			*spuptr++=af;
			break;
		case spu_vol_left:	
			if (debug_sound) printf("Main volume left %d\n",af);
			break;
		case spu_vol_right:	
			if (debug_sound) printf("Main volume right %d\n",af);
			break;
/*		case sp0:if (debug_sound) printf("sp0 written %4.4x\n",af);break;
		case sp1:if (debug_sound) printf("sp1 written %4.4x\n",af);break;
		case sp2:if (debug_sound) printf("sp2 written %4.4x\n",af);break;*/
/*Root counters stuff*/
		case t0_count: 
		case t0_mode:
		case t0_target:
		case t1_count: 
		case t1_mode:
		case t1_target:
		case t2_count: 
		case t2_mode:
		case t2_target:
			set_root_counter(hrn,af);
			break;

	}
	HRM16(hrn)=af;
}

uint8 hread_byte(uint16 hrn) {
	uint8 t;
	if (debug_hre) printf("Reading byte HR %4.4x\n",hrn);
	switch(hrn) {
	case cd_reg0:
		return(cdrom_read_reg0());
	case cd_reg1:
		return(cdrom_read_reg1());
	case cd_reg2:
		return(cdrom_read_reg2());
	case cd_reg3:
		return(cdrom_read_reg3());
	default: 
		t=HRM(hrn);
		break;
	}	
	return(t);
}

void hwrite_byte(uint8 af,uint16 hrn) {
	if (debug_hwr) printf("putting %2.2x in HR %4.4x\n",af,hrn);
	switch(hrn)
		{
		case cd_reg0:
			cdrom_write_reg0(af);
			return;
		case cd_reg1:
			cdrom_write_reg1(af);
			return;
		case cd_reg2:
			cdrom_write_reg2(af);
			return;
		case cd_reg3: 
			cdrom_write_reg3(af);
			return;
		default:
			HRM(hrn)=af;
			break;
		}
}

void hwreg(uint32 what,uint16 hf,long i) {
	switch(i) {
	case 0: hwrite_byte(what&0xff,hf);break;
	case 1: hwrite_word(what&0xffff,hf);break;
	case 2: hwrite(what,hf);break;
	}
}

uint32 hrreg(uint16 hf,long i) {
switch(i) {
	case 0: return(hread_byte(hf));
	case 1: return(hread_word(hf));
//	case 2: return(hread(hf));
	}
return(hread(hf));
}

