#include "sope.h"
#include "interpreter.h"
#include "cpu_diff.h"
#include "memio.h"
#include "gte.h"
#include "pad.h"
#include "syscall.h"
#include "hwr.h"

#include "cdrom.h"

uint32 bra_adr_b;
int interEnable = 0;

void R_ADD()  { ctx.gpr[rd]=ctx.gpr[rs]+ctx.gpr[rt]; }
void R_ADDU() { ctx.gpr[rd]=ctx.gpr[rs]+ctx.gpr[rt]; }
void R_ADDI() { ctx.gpr[rt]=ctx.gpr[rs]+ simm; }
void R_ADDIU(){ ctx.gpr[rt]=ctx.gpr[rs]+ simm; }
void R_SUBU() { ctx.gpr[rd]=ctx.gpr[rs]-ctx.gpr[rt]; }
void R_SUB()  { ctx.gpr[rd]=ctx.gpr[rs]-ctx.gpr[rt]; }

void R_BNE(){
if (ctx.gpr[rs]!=ctx.gpr[rt]) {
#include "interpreter_bra.h"
}
}

void R_BEQ(){
if (ctx.gpr[rs]==ctx.gpr[rt]) {
#include "interpreter_bra.h"
}
}

void R_BGTZ(){
if (SR[rs]>0) {
#include "interpreter_bra.h"
}
}
void R_BLEZ(){
if (SR[rs]<=0) {
#include "interpreter_bra.h"
}
}

void R_BLTZ()	{
if (SR[rs]<0) {
#include "interpreter_bra.h"
}
}

void R_BGEZ()	{
if (SR[rs]>=0) {
#include "interpreter_bra.h"
}
}
void R_BLTZAL(){
ctx.gpr[31]=ctx.pc0+8;
if (SR[rs]<0) {
#include "interpreter_bra.h"
}
}

void R_BGEZAL()   {
ctx.gpr[31]=ctx.pc0+8;
if (SR[rs]>=0) {
#include "interpreter_bra.h"
}
}

void C0_MFC()	{ ctx.gpr[rt]=ctx.cp0[rd]; }
void C0_MTC()	{ ctx.cp0[rd]=ctx.gpr[rt]; 
/*printf ("mtc[%d]=%x\n",rd,ctx.gpr[rt]);
if (rd == 12 && (ctx.gpr[rt]&2))
    {
    ctx.cp0[rd] = ctx.gpr[rt] & 0xfffffffd;
    interEnable = 1;
    }*/
}
void C0_RFE()	{ ctx.cp0[12]|=1; }

void R_DIV()  { r_div(rs,rt,0); }
void R_DIVU() { r_div(rs,rt,1); }

void R_MULU() {
/*Mult stuff by Duddie*/
unsigned int u1,u2;
unsigned int a,b,c,d,bd,ad,cb,ac;
unsigned int mid,mid2;
unsigned int carry_mid = 0;

u1=ctx.gpr[rs];u2=ctx.gpr[rt];
a = (u1>>16) & 0xffff;b = u1 & 0xffff;c = (u2>>16) & 0xffff;d = u2 & 0xffff;
bd = b*d;ad = a*d;cb = c*b;ac = a*c;mid = ad+cb;
if(mid<ad || mid<cb) carry_mid = 1;
mid2 = mid +((bd>>16)&0xffff);
if (mid2<mid || mid2 < ((bd>>16)&0xffff)) carry_mid +=1;
ctx.LO = (bd & 0xffff) | ((mid2 & 0xffff) << 16);
ctx.HI = ac + (carry_mid<<16) + ((mid2 >> 16) &  0xffff);
}

void R_MUL() {
/*Mult stuff by Duddie*/
int signed v1,v2;
unsigned int a,b,c,d,bd,ad,cb,ac;
unsigned int mid,mid2;
unsigned int carry_mid = 0;
int neg_sign = 0;

v1=SR[rs];v2=SR[rt];
if (v1<0) {v1 = -v1;neg_sign = 1;}
if (v2<0) {v2 = -v2;neg_sign = ! neg_sign;}
a = (v1>>16) & 0xffff;b = v1 & 0xffff;
c = (v2>>16) & 0xffff;d = v2 & 0xffff;
bd = b*d;ad = a*d;cb = c*b;ac = a*c;mid = ad+cb;
if(mid<ad || mid<cb) carry_mid = 1;
mid2 = mid +((bd>>16)&0xffff);
if (mid2<mid || mid2 < ((bd>>16)&0xffff)) carry_mid +=1;
ctx.LO = (bd & 0xffff) | ((mid2 & 0xffff) << 16);
ctx.HI = ac + (carry_mid<<16) + ((mid2 >> 16) &  0xffff);
if (neg_sign) {
	ctx.LO = ~ctx.LO;ctx.HI = ~ctx.HI;
	ctx.LO +=1;if (ctx.LO == 0) ctx.HI+=1;
}
}

void R_MFHI() { ctx.gpr[rd]=ctx.HI; }
void R_MFLO() { ctx.gpr[rd]=ctx.LO; }
void R_MTHI() { ctx.HI=ctx.gpr[rs]; }
void R_MTLO() { ctx.LO=ctx.gpr[rs]; }

void R_LWC2(){ ctx.gte1[rt]=rm_32(ctx.gpr[rs]+simm); }
void R_SWC2(){ wm_32(ctx.gte1[rt],ctx.gpr[rs]+simm); }
void R_GTE() { gte_opcode(opc); }

void R_J(){
#include "interpreter_jmp.h"
}

void R_JR(){
#define __DYNJ__ 1
#include "interpreter_jmp.h"
#undef __DYNJ__
}

void R_JAL(){
#define __LINJ__ 1
#include "interpreter_jmp.h"
#undef __LINJ__
}

void R_JALR(){
#define __LINJ__ 1
#define __DYNJ__ 1
#include "interpreter_jmp.h"
#undef __DYNJ__
#undef __LINJ__
}

void R_NOR() 	{ ctx.gpr[rd]=(ctx.gpr[rs] | ctx.gpr[rt])^(0xFFFFFFFF); }
void R_OR()  	{ ctx.gpr[rd]=ctx.gpr[rs] | ctx.gpr[rt]; }
void R_XOR() 	{ ctx.gpr[rd]=ctx.gpr[rs] ^ ctx.gpr[rt]; }
void R_AND() 	{ ctx.gpr[rd]=ctx.gpr[rs] & ctx.gpr[rt]; }
void R_ORI() 	{ ctx.gpr[rt]=ctx.gpr[rs] | imm; }
void R_XORI()	{ ctx.gpr[rt]=ctx.gpr[rs] ^ imm; }
void R_ANDI()	{ ctx.gpr[rt]=ctx.gpr[rs] & imm; }

void R_LUI(){ctx.gpr[rt]=(imm << 16);}

void R_LW(){  if (rt==0) return;ctx.gpr[rt]=rm_32(ctx.gpr[rs]+simm); }
void R_LHU(){ if (rt==0) return;ctx.gpr[rt]=rm_16u(ctx.gpr[rs]+simm); }
void R_LBU(){ if (rt==0) return;ctx.gpr[rt]=rm_8u(ctx.gpr[rs]+simm); }
void R_LH(){  if (rt==0) return;SR[rt]=rm_16s(ctx.gpr[rs]+simm); }
void R_LB(){  if (rt==0) return;SR[rt]=rm_8s(ctx.gpr[rs]+simm); }

void R_SW(){wm_32(ctx.gpr[rt],ctx.gpr[rs]+simm); }
void R_SH(){wm_16(ctx.gpr[rt],ctx.gpr[rs]+simm); }
void R_SB(){wm_8( ctx.gpr[rt],ctx.gpr[rs]+simm); }

void R_XX() { ciao("Invalid Opcode"); }
void R_BREAK() { ciao("Break Occured!"); }

void R_SYSC() {
ctx.cp0[12]&=0xfffffffe;
ctx.cp0[13]=0x20;
ctx.cp0[14]=ctx.pc0;
ctx.pc0=0x80-4;
}

void R_SLL(){
if (opc==0) return;
ctx.gpr[rd]=(ctx.gpr[rt] << re);
}

void R_SRL(){ ctx.gpr[rd]=(ctx.gpr[rt] >> re); }
void R_SRA(){ SR[rd]=(SR[rt] >> re); }

void R_SLLV(){ ctx.gpr[rd]=(ctx.gpr[rt] << ctx.gpr[rs]); }
void R_SRLV(){ ctx.gpr[rd]=(ctx.gpr[rt] >> ctx.gpr[rs]); }
void R_SRAV(){ SR[rd]=(SR[rt] >> ctx.gpr[rs]); }

void R_SLT(){ if (SR[rs]<SR[rt]) ctx.gpr[rd]=1; else ctx.gpr[rd]=0; }
void R_SLTU(){ if (ctx.gpr[rs]<ctx.gpr[rt]) ctx.gpr[rd]=1; else ctx.gpr[rd]=0; }
void R_SLTI(){ if (SR[rs]<(simm)) ctx.gpr[rt]=1; else ctx.gpr[rt]=0; }
void R_SLTIU(){
uint32 simm_se;
simm_se=simm;
if (ctx.gpr[rs]<simm_se) ctx.gpr[rt]=1; else ctx.gpr[rt]=0;
}

void R_LWL(){ r_wrl(ctx.gpr[rs]+simm,rt,0); }
void R_LWR(){ r_wrl(ctx.gpr[rs]+simm,rt,1); }
void R_SWL(){ r_wrl(ctx.gpr[rs]+simm,rt,2); }
void R_SWR(){ r_wrl(ctx.gpr[rs]+simm,rt,3); }

void (*regimm[32])(void)={  R_BLTZ,R_BGEZ,R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,
                            R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,
                            R_BLTZAL,R_BGEZAL,R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,
                            R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,R_XX};

void (*cop0[32])(void)={C0_MFC,R_XX,R_XX,R_XX,C0_MTC,R_XX,R_XX,R_XX,
                        R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,
                        C0_RFE,R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,
                        R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,R_XX};

void R_REGM()   {regimm[rt]();}
void R_CP0()    {cop0[rs]();}

void (*special[64])(void)={ R_SLL,R_XX,R_SRL,R_SRA,R_SLLV,R_XX,R_SRLV,R_SRAV,
                            R_JR,R_JALR,R_XX,R_XX,R_SYSC,R_BREAK,R_XX,R_XX,
                            R_MFHI,R_MTHI,R_MFLO,R_MTLO,R_XX,R_XX,R_XX,R_XX,
                            R_MUL,R_MULU,R_DIV,R_DIVU,R_XX,R_XX,R_XX,R_XX,
                            R_ADD,R_ADDU,R_SUB,R_SUBU,R_AND,R_OR,R_XOR,R_NOR,
                            R_XX,R_XX,R_SLT,R_SLTU,R_XX,R_XX,R_XX,R_XX,
                            R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,
                            R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,R_XX};

void R_SPEC()   {special[opc&63]();}


void (*normal[64])(void)={  R_SPEC,R_REGM,R_J,R_JAL,R_BEQ,R_BNE,R_BLEZ,R_BGTZ,
                            R_ADDI,R_ADDIU,R_SLTI,R_SLTIU,R_ANDI,R_ORI,R_XORI,R_LUI,
                            R_CP0,R_XX,R_GTE,R_XX,R_XX,R_XX,R_XX,R_XX,
                            R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,R_XX,
                            R_LB,R_LH,R_LWL,R_LW,R_LBU,R_LHU,R_LWR,R_XX,
                            R_SB,R_SH,R_SWL,R_SW,R_XX,R_XX,R_SWR,R_XX,
                            R_XX,R_XX,R_LWC2,R_XX,R_XX,R_XX,R_XX,R_XX,
                            R_XX,R_XX,R_SWC2,R_XX,R_XX,R_XX,R_XX,R_XX};

void Fetch_Instruction(){
    opc=getopc(ctx.pc0);
    normal[opc>>26]();
//    if (dass) dasm2();
    ctx.nstp++;
    ctx.pc0+=4;
}


void handle_flow() {
	cd_misc();
if (cd_int)
    {
    cd_int=0;
//    if (interEnable) {
    if (ctx.cp0[12]&1) {
    ctx.cp0[13]=0x400;
    if (((ctx.cp0[13] & ctx.cp0[12] & 0x400)==0x400))
        {
        ctx.cp0[12]&=0xfffffffe;
        ctx.cp0[14]=ctx.pc0;
        ctx.pc0=0x80;
        interEnable = 0;
        }
    }
    }
if (enable_vsync)
    if ((ctx.nstp>ncyc) && (ctx.cp0[12]&1))
    {
    redrawframe();

    HRM32(i_reg)|=1;
    ctx.cp0[13]=0x400;
    ctx.cp0[12]&=0xfffffffe;
    ctx.cp0[14]=ctx.pc0;
    ctx.pc0=0x80;
    ctx.nstp=0;
    get_pad_buf();
    }


/*	if (cdint_enable) {
		if (ctx.cp0[12]&1) {
			ctx.cp0[13]=0x400;
			ctx.cp0[12]&=0xfffffffe;
			ctx.cp0[14]=ctx.pc0;
			ctx.pc0=0x80;
			HRM(i_reg)|=4;
			cdint_enable=0;
		}
	}
	if (enable_vsync) {
		if ((ctx.nstp>ncyc) && (ctx.cp0[12]&1)) {
				redrawframe();
				HRM(i_reg)|=1;
				ctx.cp0[13]=0x400;
				ctx.cp0[12]&=0xfffffffe;
				ctx.cp0[14]=ctx.pc0;
				ctx.pc0=0x80;
				ctx.nstp=0;
				get_pad_buf();
		}
	}
*/
	switch (ctx.pc0) {
	case 0xffffffff: ciao("Break Occured");
	case 0x00000000: ciao("GROS PROBLEME!!!!");
	case 0x000000a0: syscall_a0(ctx.gpr[9]);break;
	case 0x000000b0: syscall_b0(ctx.gpr[9]);break;
	case 0x000000c0: syscall_c0(ctx.gpr[9]);break;
	}

	Fetch_Instruction(); 
}


