#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include "opengl.h"
#include "keyboard.h"
#include "vram_macros.h"

int cur_width,cur_height;

Config_List ogl_config_list={0,NULL};

Renderer ogl_renderer = {
	"OpenGL renderer",
	&ogl_config_list,
	ogl_start,
	ogl_init,
	ogl_exit,
	ogl_open_window,
    ogl_close_window,
    ogl_FSline,
    ogl_GSline,
    ogl_FSpoly3,
    ogl_GSpoly3,
    ogl_FSpoly4,
    ogl_GSpoly4,
    ogl_FSTpoly3,
    ogl_GSTpoly3,
    ogl_FSTpoly4,
    ogl_GSTpoly4,
    ogl_render_sprite,
    ogl_blockfill,
    ogl_rectangle,
    ogl_clear_texture_cache,
    ogl_render_frame,
    ogl_set_scissor,
    ogl_set_blending,
    ogl_change_title,
	ogl_key_pressed,
	ogl_get_key
};

Renderer *get_renderer_info() {
	return &ogl_renderer;
}

Display *Dsp;
XEvent  E;
Window Wnd;
GLXContext glctx;

void open_context() {
	int attrib[] = {GLX_RGBA,
			        GLX_RED_SIZE, 1,
					GLX_GREEN_SIZE, 1,
					GLX_BLUE_SIZE, 1,
					GLX_DOUBLEBUFFER,
					None };
	int scrnum;
	XSetWindowAttributes attr;
	unsigned long mask;
	Window root;
	XVisualInfo *visinfo;
#define w g_width
#define h g_height
    Dsp=XOpenDisplay(NULL);
	scrnum = DefaultScreen( Dsp );
	root = RootWindow( Dsp, scrnum );
	visinfo = glXChooseVisual( Dsp, scrnum, attrib );
	if (!visinfo) {
		printf("Error: couldn't get an RGB, Double-buffered visual\n");
		exit(1);
	}

	/* window attributes */
	attr.background_pixel = 0;
	attr.border_pixel = 0;
	attr.colormap = XCreateColormap( Dsp, root, visinfo->visual, AllocNone);
	attr.event_mask = StructureNotifyMask | ExposureMask | KeyPressMask;
	mask = CWBackPixel | CWBorderPixel | CWColormap | CWEventMask;

	Wnd = XCreateWindow( Dsp, root, 0, 0, w, h,
						 0, visinfo->depth, InputOutput,
						 visinfo->visual, mask, &attr );
	{
		XSizeHints Hints;
		XWMHints WMHints;
		Hints.flags=PSize|PMinSize|PMaxSize;
		Hints.base_width=w;
		Hints.base_height=h;
		Hints.min_width=0;
		Hints.max_width=1024;
		Hints.min_height=0;
		Hints.max_height=768;
		WMHints.input=True;WMHints.flags=InputHint;
		XSetWMHints(Dsp,Wnd,&WMHints);
		XSetWMNormalHints(Dsp,Wnd,&Hints);
		XStoreName(Dsp,Wnd,"Sope");
		XSelectInput(Dsp,Wnd,FocusChangeMask|ExposureMask|KeyPressMask|KeyReleaseMask|StructureNotifyMask);
		XMapRaised(Dsp,Wnd);
		XClearWindow(Dsp,Wnd);
		XAutoRepeatOff(Dsp);
		XWindowEvent(Dsp,Wnd,ExposureMask,&E);
	}
	glctx = glXCreateContext( Dsp, visinfo, NULL, True );

	glXMakeCurrent( Dsp, Wnd, glctx );
	XMapWindow(Dsp,Wnd);
    glNewList(1,GL_COMPILE);
}

void ogl_change_title(char *title) {
    XStoreName(Dsp,Wnd,title);
}

void ogl_start(gpu_state *gs,uint32 *vr) {
    int c[2];
    
    gps=gs;
    vram=vr;

	g_width=640;
	g_height=480;
    open_context();
    glGetIntegerv(GL_MAX_VIEWPORT_DIMS,c);
    max_vpw=c[0];
    max_vph=c[1];
    glDisable(GL_BLEND);
    glDisable(GL_SCISSOR_TEST);
    glDisable(GL_ALPHA_TEST);
    glAlphaFunc(GL_NOTEQUAL,0.0);
    glScissor(0,0,1024,512);
    gps->oxi=0;gps->oyi=0;gps->oxf=1024;gps->oyf=512;
}

void ogl_init() {
}

void ogl_exit() {

	printf("# Sope # Shutting down renderer...");
	glXDestroyContext(Dsp,glctx);
	XDestroyWindow(Dsp,Wnd);
	XAutoRepeatOn(Dsp);
	XCloseDisplay(Dsp);
	printf("OK\n");
}

void ogl_open_window(unsigned int w,unsigned int h) {
	char titre[256];

	cur_width=w;
	cur_height=h;
	sprintf(titre,"Sope -- %ux%u --",cur_width,cur_height);
	ogl_change_title(titre);
}

void ogl_close_window() {
}

void ogl_render_frame() {
    glEndList();
    glViewport(0,adj_h(cur_height-512),adj_w(glw),adj_h(glh));
    glCallList(1);
    glDeleteLists(1,1);
    glNewList(1,GL_COMPILE);
    glXSwapBuffers(Dsp,Wnd);
}

void ogl_set_scissor() {
sint32 ncx,ncy,ncw,nch;

if ((cur_width==0) || (cur_height==0)) return;
ncx=adj_w(gps->clip_x);
ncy=adj_h(gps->clip_y);
ncw=adj_w(gps->clip_w)+1-ncx;
nch=adj_h(gps->clip_h)+1-ncy;
glScissor(ncx,ncy,ncw,nch);
}

void ogl_set_blending() {
if (screen_mode==0xff) return;
switch(gps->abr) {
	case 0: glBlendColorEXT(0.0,0.0,0.0,0.5);
			glBlendFunc(GL_CONSTANT_ALPHA,GL_CONSTANT_ALPHA);
		    glBlendEquationEXT(GL_FUNC_ADD_EXT);
			break;
	case 1: glBlendFunc(GL_ONE,GL_ONE);
			glBlendEquationEXT(GL_FUNC_ADD_EXT);
		    break;          
	case 2: glBlendFunc(GL_ONE,GL_ONE);
		    glBlendEquationEXT(GL_FUNC_REVERSE_SUBTRACT_EXT);
	        break;      
	case 3: glBlendColorEXT(0.0,0.0,0.0,0.25);
	        glBlendFunc(GL_CONSTANT_ALPHA,GL_ONE);
	        glBlendEquationEXT(GL_FUNC_ADD_EXT);
	        break;      
}
}
XEvent E;

int ogl_key_pressed(){
    XCheckWindowEvent(Dsp,Wnd,KeyPressMask|KeyReleaseMask,&E);
    if (E.type==KeyPress) return PRESSED_KEY;
    if (E.type==KeyRelease) return RELEASED_KEY;
    return NO_KEY;
}

uint16 ogl_get_key(){
    KeySym key;

    key=XLookupKeysym((XKeyEvent *)&E,0);
    switch(key){
        case XK_Escape: return KEY_ESCAPE;
        case XK_Up:     return KEY_UP;
        case XK_Down:   return KEY_DOWN;
        case XK_Left:   return KEY_LEFT;
        case XK_Right:  return KEY_RIGHT;
        default:        return key;
    }
}

