#include <malloc.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include "soft.h"
#include "keyboard.h"

static int cur_width,cur_height;
static uint16 dsp_keymap[km_s];

Display *Dsp;
XEvent	E;
Screen	*Scr;
Window	Wnd;
GC	DefaultGC;
uint8	*XBuf;
uint16	*XBuf16;
uint32	*XBuf32;
XImage	*ximage;
int colordepth=0;

static int dsp_init();
static void dsp_exit();
static void dsp_open_window(unsigned int w,unsigned int h);
static void dsp_close_window();
static void dsp_change_title(char *title);
static void dsp_render_frame(uint32 *image, int start, int endw);
static int dsp_key_pressed();
static uint16 dsp_get_key();

static void dsp_fillkeymap();

DisplayStruct Dsp_X11 = {
	dsp_init,
	dsp_exit,
	dsp_open_window,
	dsp_close_window,
	dsp_change_title,
	dsp_render_frame,
	dsp_key_pressed,
	dsp_get_key,
};


static int dsp_init() {
	gps->oxi=0;gps->oyi=0;gps->oxf=1024;gps->oyf=512;
	Dsp = XOpenDisplay(NULL);
	if (Dsp==NULL) return -1;
	XCloseDisplay(Dsp);
	dsp_fillkeymap();
	return 0;
}

static void dsp_open_window(unsigned int w,unsigned int h) {
	long	unsigned White,Black;
	int i,attr_mask;
	XSetWindowAttributes attr;
	Colormap cmap = 0;
	XColor colors[256];
	char titre[256];

	cur_width=w;
	cur_height=h;
	Dsp=XOpenDisplay(NULL);
	Scr=DefaultScreenOfDisplay(Dsp);
	White=WhitePixelOfScreen(Scr);
	colordepth=DefaultDepthOfScreen(Scr);
	Black=BlackPixelOfScreen(Scr);
	DefaultGC=DefaultGCOfScreen(Scr);
	attr.event_mask = KeyPressMask | KeyReleaseMask;
	attr.background_pixel = Black;
	attr.border_pixel = White;
	attr.backing_store = WhenMapped;
	attr_mask = CWBackPixel | CWEventMask | CWBorderPixel | CWBackingStore;

	Wnd=XCreateWindow(Dsp,RootWindowOfScreen(Scr),0,0,w,h,0,colordepth,
			  InputOutput,DefaultVisualOfScreen(Scr),attr_mask,&attr);
	{
		XSizeHints Hints;
		XWMHints WMHints;
		Hints.flags=PSize|PMinSize|PMaxSize;
		Hints.min_width=Hints.max_width=Hints.base_width=w;
		Hints.min_height=Hints.max_height=Hints.base_height=h;
		WMHints.input=True;WMHints.flags=InputHint;
		XSetWMHints(Dsp,Wnd,&WMHints);
		XSetWMNormalHints(Dsp,Wnd,&Hints);
		sprintf(titre,"Sope -- %ux%u --",w,h);
		XStoreName(Dsp,Wnd,titre);
		XSelectInput(Dsp,Wnd,FocusChangeMask|ExposureMask|KeyPressMask|KeyReleaseMask);
		XMapRaised(Dsp,Wnd);
		XClearWindow(Dsp,Wnd);
//		XAutoRepeatOff(Dsp);
		XWindowEvent(Dsp,Wnd,ExposureMask,&E);}

	XBuf=(uint8 *)malloc((colordepth/8)*w*h);
	XBuf16=(uint16 *)XBuf;
	XBuf32=(uint32 *)XBuf;
	if (colordepth==8) {
		cmap = XCreateColormap (Dsp,Wnd,DefaultVisualOfScreen(Scr),True);
		XSetWindowColormap (Dsp,Wnd, cmap);
		for (i = 0; i < 256; i++) {
			colors[i].red = colors[i].green = colors[i].blue = i<<8;
			colors[i].pixel = i;
			colors[i].flags = DoRed | DoGreen | DoBlue;
		}
		XStoreColors (Dsp, cmap, colors, 256);
	}
	if (colordepth==32)
		ximage=XCreateImage(Dsp,DefaultVisualOfScreen(Scr),24,ZPixmap,0,XBuf,w,h,32,0);
	else
		ximage=XCreateImage(Dsp,DefaultVisualOfScreen(Scr),colordepth,ZPixmap,0,XBuf,w,h,32,0);
	return;
}
static void dsp_close_window() {
	XDestroyWindow(Dsp,Wnd);
//	XAutoRepeatOn(Dsp);
	XCloseDisplay(Dsp);
	free(XBuf);
}

static void dsp_exit() {
	XDestroyWindow(Dsp,Wnd);
	XAutoRepeatOn(Dsp);
	XCloseDisplay(Dsp);
	free(XBuf);
}
static void dsp_change_title(char *title) {
	XStoreName(Dsp,Wnd,title);
}

static void dsp_render_frame(uint32 *image, int start, int endw) {
	int unsigned r,g,b;
	int unsigned i,j,k,l;
	uint32 *ll;
	uint16 cc;

	ll=image+start;
	k=0;
	switch (colordepth) {
	case 32:
	for (j=0;j<cur_height;j++) {
		for (i=0;i<cur_width;i++) {
			l=*ll;
			r=l&0xff;
			g=(l>>8)&0xff;
			b=(l>>16)&0xff;
			XBuf32[k]=b+256*g+65536*r;
			k++;
			ll++;
		}
		ll+=(endw-cur_width);
	}
	break;
	case 24:
		for (j=0;j<cur_height;j++) {
			for (i=0;i<cur_width;i++) {
				l=ll[j*endw+i];
				XBuf[k]=(l>>16)&0xff;
				XBuf[k+1]=(l>>8)&0xff;
				XBuf[k+2]=l&0xff;
				k+=3;
			}
		}
	break;
	case 16:
		for (j=0;j<cur_height;j++) {
			for (i=0;i<cur_width;i+=2) {
				int l1=ll[0], l2=ll[1]; // byte order affects whether it is 0,1 or 1,0
				XBuf32[k] = 
					((l1<<8)&(31<<11))  |
					((l1>>19)&31)	   |
					((l1>>5)&(31<<6))   |
					((l2<<24)&(31<<27)) |
					((l2>>3)&(31<<16))  |
					((l2<<11)&(31<<22));
				k++;
				ll+=2;
			}
			ll+=(endw-cur_width);
		}
		break;
	case 8:
		for (j=0;j<cur_height;j++) {
			for (i=0;i<cur_width;i++) {
				l=ll[j*endw+i];
				cc=l&0xff;
				cc+=(l>>8)&0xff;
				cc+=(l>>16)&0xff;
				cc/=3;
				XBuf[k]=cc;
				k++;
			}
		}
		break;
	}
	XPutImage(Dsp,Wnd,DefaultGC,ximage,0,0,0,0,cur_width,cur_height);
	XFlush(Dsp);
}

static int dsp_key_pressed(){
	XCheckWindowEvent(Dsp,Wnd,KeyPressMask|KeyReleaseMask,&E);
	if (E.type==KeyPress) return PRESSED_KEY;
	if (E.type==KeyRelease) return RELEASED_KEY;
	return NO_KEY;
}

static uint16 dsp_get_key(){
	KeySym key;

	key=XLookupKeysym((XKeyEvent *)&E,0);
	switch(key){
		case XK_Escape: return KEY_ESCAPE;
		case XK_Up:		return KEY_UP;
		case XK_Down:	return KEY_DOWN;
		case XK_Left:	return KEY_LEFT;
		case XK_Right:	return KEY_RIGHT;
		case XK_F1:	return KEY_F1;
		case XK_F2:	return KEY_F2;
		case XK_F3:	return KEY_F3;
		case XK_F4:	return KEY_F4;
		case XK_F5:	return KEY_F5;
		case XK_F6:	return KEY_F6;
		case XK_F7:	return KEY_F7;
		case XK_F8:	return KEY_F8;
		case XK_F9:	return KEY_F9;
		case XK_F10:	return KEY_F10;
		case XK_F11:	return KEY_F11;
		case XK_F12:	return KEY_F12;
		default:	return key;
	}
}


static void dsp_fillkeymap() {
	int i;

	for (i=0; i<km_s; i++) {
	    dsp_keymap[i] = i;
	}
}