#include <string.h>
#include "soft.h"
#include "keyboard.h"
#include "vram_macros.h"

Config_Param soft_config_params[1]={
	{boolean,"Transparency",1}
};

Config_List soft_config_list={1,soft_config_params};

Renderer soft_renderer = {
	"Custom software renderer",
	&soft_config_list,
	soft_start,
	soft_init,
	soft_exit,
	soft_open_window,
	soft_close_window,
	soft_FSline,
	soft_GSline,
	soft_FSpoly3,
	soft_GSpoly3,
	soft_FSpoly4,
	soft_GSpoly4,
        soft_FSTpoly3,
	soft_GSTpoly3,
	soft_FSTpoly4,
        soft_GSTpoly4,
        soft_render_sprite,
        soft_blockfill,
	soft_rectangle,
	soft_clear_texture_cache,
	soft_render_frame,
	soft_set_scissor,
	soft_set_blending,
	soft_change_title,
	soft_key_pressed,
	soft_get_key
};

DisplayStruct Dsp_X11;
DisplayStruct Dsp_SDL;
DisplayStruct Dsp_AA;
DisplayStruct Dsp_VGA;

DisplayStruct *Displays[] = {
&Dsp_X11,
&Dsp_VGA,
&Dsp_AA,
&Dsp_SDL,
NULL // end of displays
};

DisplayStruct Dsp;
static int fullscreen = 0;
#define fullscrW	640
#define fullscrH	480
static int realw,realh;

void soft_init() {
	int i;
	for (i=0; Displays[i]!=NULL; i++)
	    if (!Displays[i]->init()) {
		Dsp = *Displays[i];
		return;
	    }
	printf ("no Dsp avail\n");
	exit(1);
}

void soft_exit() {
	Dsp.exit();
}

void soft_open_window(unsigned int w, unsigned int h) {
	realw = w;
	realh = h;
	if (fullscreen) {
	    Dsp.open_window(fullscrW,fullscrH);
	}
	else Dsp.open_window(w,h);
}

void soft_close_window() {
	Dsp.close_window();
}

void soft_change_title(char *title) {
	Dsp.change_title(title);
}

void soft_render_frame() {
	if (!fullscreen) Dsp.render_frame(vram,gps->dofs_y*1024+gps->dofs_x,1024);
	else {
	    uint32 tmp[fullscrW*fullscrH];
	    soft_scale(vram,realw,realh,tmp,fullscrW,fullscrH);
	    Dsp.render_frame(tmp,0,fullscrW);
	}
}

int soft_key_pressed() {
	return Dsp.key_pressed();
}

uint16 soft_get_key() {
	return Dsp.get_key();
}

Renderer *get_renderer_info() {
	return &soft_renderer;
}

void soft_start(gpu_state *gs,uint32 *vr) {
	gps=gs;
	vram=vr;
}

/*Color Modulation*/
uint32 moco(uint32 cc,uint32 c) {
	uint32 cr,cg,cb;
	cr=((cc&255)*(c&255))>>7;
	cg=(((cc>>8)&255)*((c>>8)&255))>>7;
	cb=(((cc>>16)&255)*((c>>16)&255))>>7;
	if (cr>0xff) cr=0xff;
	if (cg>0xff) cg=0xff;
	if (cb>0xff) cb=0xff;
	return cr+(cg<<8)+(cb<<16)+(cc&0xFF000000);
}

inline void draw_sumcol(uint16 x,uint16 y,int p,uint32 c2)
{
	uint32 t,mask,c1;

	c1=vram[y*1024+x];
// if we aren't supposed to write to masked pixels, and this pixel
// is masked, then don't write
// if the color to write/append is black without STP set, then
// it is transparent no matter what
//	if ((_me_&c1) || (!p && !c2))
	if ((gps->_me_&c1) || !c2)
		return;
	if (!p || ((c2&0xFFFFFF) && ((c2&0xFF000000)==0))) {
		vram[y*1024+x]=(c2&0xFFFFFF)|gps->_md_|(c1&0xFF000000);
		return;
	}
/*	if (p & (y==0xFF000000)) return _md_;*/
// if we are not in transparent mode, or the color to write/append is
// set to opaque, then write
/*	if( !p || ((y&0xFF000000) && (y&0xFFFFFF)))
//	if( !p || ((y&0xFF000000)==0)
		return (y&0xFFFFFF)|_md_|(x&0xFF000000);*/
	switch(gps->abr) {
	case 0:
		t=((c1>>1)&0x7f7f7f)+((c2>>1)&0x7f7f7f);
		break;
 
	case 3:
		//t=0;
		//for (i=0;i<3;i++) {
		//	c=(x>>(8*i))&0xff;d=(y>>(8*i))&0xff;
		//	tc=c+(d>>2);
		//	if (tc>0xff) tc=0xff;
		//	if (tc<0)	 tc=0;
		//	t+=(tc<<(8*i));
		//}
		//break;
		c2=(c2>>2)&0x3f3f3f;
		// fall through and let case 1 handle it

	case 1:
/*		t2=0;
		for (i=0;i<3;i++) {
			c=(x>>(8*i))&0xff;d=(y>>(8*i))&0xff;
			tc=c+d;
			if (tc>0xff) tc=0xff;
			t2+=(tc<<(8*i));
		}*/
		t=c1+c2;
		mask=(c1^c2^t)&0x1010100;
		t=((t-mask)|((mask>>8)*255))&0xFFFFFF;
		//assert(t==t2);
		break;
 
	case 2:
		//t2=0;
		//for (i=0;i<3;i++) {
		//	c=(x>>(8*i))&0xff;d=(y>>(8*i))&0xff;
		//	tc=c-d;
		//	if (tc>0xff) tc=0xff;
		//	if (tc<0)	 tc=0;
		//	t2+=(tc<<(8*i));
		//}
		t=c1-c2;
		mask=(c1^c2^t)&0x1010100;
		t=((t+mask)&(~((mask>>8)*255)))&0xFFFFFF;
		//assert(t==t2);
		break;
		default:t=0;break; 
	}
	vram[1024*y+x]=(t|gps->_md_|(c1&0xFF000000));
	return;
}

void soft_set_scissor() {
}

void soft_set_blending() {
}

static int lpos[514],rpos[514];
static int top,bottom;
static uint32 colorarray[256];

void sd_initcc(texture *tex,int m,uint32 c)
{
	int i;

	if( tex->tx_pal == 2 )
		return;
//	if (m) 
		for( i=0; i<16+240*tex->tx_pal; i++ )
			colorarray[i]=vram[tex->cluy*1024+tex->clux+i];
/*NOTE: possible speed-up for FST: blend the palette*/
/*	else
		for( i=0; i<16+240*tex->tx_pal; i++ )
			colorarray[i]=moco(c,vram[tex->cluy*1024+tex->clux+i]);
*/
}
#define MAX(a,b) ((a>b)?a:b)
#define MIN(a,b) ((a<b)?a:b)

/*inline int MAX(int a,int b)
{
	if( a > b )
		return a;
	return b;
}

inline int MIN(int a,int b)
{
	if( a < b )
		return a;
	return b;
}
*/
static void sd_init()
{
	int i;
	for( i=0; i<514; i++ )
	{
		lpos[i]=1024;
		rpos[i]=0;
	}
//	memset(rpos,0,2056);
	top=512;
	bottom=0;
}

static inline int computex2(int y,int toprx,int topry)
{
	int x=rpos[y+1]-1;
	if( topry == y )
		x=MIN(x,toprx-1);
	x=MIN(x,rpos[y+2]);
	if( y > top )
		x=MIN(x,rpos[y]);
	return MIN(gps->oxf-1,x);
}

static void sd_addpoint(int x,int y)
{
	/* we include the line above and below the clipping region */
	if( y >= (gps->oyi-1) && y <= gps->oyf )
	{
		if( y < top )
			top=y;
		if( y > bottom )
			bottom=y;

		if( x < lpos[y+1] )
			lpos[y+1]=x;
		if( x > rpos[y+1] )
			rpos[y+1]=x;
	}
}

static void sd_addline(int x1,int y1,int x2,int y2)
{
	if( y1==y2 )
	{
		sd_addpoint(x1,y1);
		sd_addpoint(x2,y2);
	} else if( y2 < y1 )
	{
		sd_addline(x2,y2,x1,y1);
	} else {
		int i,curx,dir;

		sd_addpoint(x1,y1);
		if( x2>x1)
			dir=-1;
		else if( x2<x1 )
			dir=1;
		else
			dir=0;
		for( i=y1+1; i<=y2; i++ )
		{
			curx=x1+((2*(i-y1)-1)*(x2-x1)-dir*(y2-y1))/(2*(y2-y1));
			sd_addpoint(curx+dir,i-1);
			sd_addpoint(curx,i);
		}
		sd_addpoint(x2,y2);
	}
}

/*   0---1
	 |  /
	 | /
	 |/
	 2

	 0: (a,b)=(0,0)
	 1: (a,b)=(255,0)
	 2: (a,b)=(0,255)
*/
void sdf_triangle(sint16 *xa,sint16 *ya,int p0,int p1,int p2,int blendP,uint32 *colors) {
#include "poly.h"
}
void sdg_triangle(sint16 *xa,sint16 *ya,int p0,int p1,int p2,int blendP,uint32 *colors) {
#define GOURAUD
#include "poly.h"
#undef GOURAUD
}
void sdft_triangle(sint16 *xa,sint16 *ya,int p0,int p1,int p2,int blendP,uint32 *colors,int mocoP,texture *tex) {
#define TEXTURED
#include "poly.h"
#undef TEXTURED
}
void sdgt_triangle(sint16 *xa,sint16 *ya,int p0,int p1,int p2,int blendP,uint32 *colors,int mocoP,texture *tex) {
#define GOURAUD
#define TEXTURED
#include "poly.h"
#undef TEXTURED
#undef GOURAUD
}

void sd_line(sint16 *xa,sint16 *ya,uint32 *c,int gourP,int blendP)
{
/*FIXME: Gouraud lines still need to be implemented*/
	int e,y,x;

	if( gourP && (c[0]==c[1]))
		gourP=0;
	if(( gourP ) && (debug_poly))
		printf("Goury line %x to %x\n",c[0],c[1]);
	if( !gourP )
		e=c[0]|0xFF000000;
	sd_init();
	sd_addline(xa[0],ya[0],xa[1],ya[1]);
	for( y=MAX(gps->oyi,top); y<=MIN(gps->oyf-1,bottom); y++ )
//		for( y=top; y<=bottom; y++ )
	{
		for( x=MAX(gps->oxi,lpos[y+1]); x<=MIN(gps->oxf-1,rpos[y+1]); x++ )
		{
			if( gourP )
			{
				e=c[0]|0xFF000000;
			}
			draw_sumcol(x,y,blendP,e);
		}
	}
}

void soft_clear_texture_cache(){

}


void soft_scale (uint32 *src, int sw, int sh,
		 uint32 *dst, int dw, int dh) {
	int x,y,i;
	float xs,ys,xx,yy;

	xs = (float)dw / sw;
	ys = (float)dh / sh;

	if ((xs==1) & (ys==1)) {
		for (y=0; y<dh; y++) {
		    for (x=0; x<dw; x++) *(dst++) = *(src++);
		    src+=1024-sw;
		}
	    return;
	    }

	yy = 0;
	for (y=0; y<sh; y++) {
	    xx = 0;
	    yy += ys;
	    if (yy >= 1) {
		for (x=0; x<sw; x++) {
		    xx += xs;
		    if (xx >= 1) {
			for (i=0; i<(int)xx; i++) {
			    int j;

			    for (j=0; j<(int)yy; j++) 
			        *(dst+(dw*j)) = *src;
			    dst++;
			}
			src++;
			xx -= (int)xx;
		    }
		    else src++;
		}
		for (i=1; i<(int)yy; i++) {
 		    dst+=dw;
		}
		yy -= (int)yy;
 		src+=1024-sw;
	    }
	    else src+=1024;
	}
}
