#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include "sope.h"
#include "sys/time.h"

#include "gui.h"
#include "renderer.h"
#include "plugins.h"

#include "sound.h"
#include "spu.h"

#include "init.h"

UserInterface console_ui;
UserInterface gtk_ui;

void print_help(int argc, char **argv) {
	printf(
"---------------------------------------------------------\n"
"Usage: %s [options] [files]\n"
"Options are:\n"
"    -v=NCYC VBLANK interrupt every NCYC*1000 instructions\n"
"    -s      Enable sound\n"
"    -p      Forces PAL mode\n"
"    -h      Displays this text\n"
"    -l      List available plug-ins\n"
"    -b      Enable bilinear filtering (OpenGL renderer)\n"
"    -d      Disable dynamic recompilation\n"
"    -n      Disable texturing\n"
"    -B      Boots CD-ROM\n",argv[0]);
}

void make_debug_flags();

void ciao(char *ph) {

	if (screen_mode!=0xff) rnd->exit();
	if (disable_sound==0)  snd->exit();
	gui->print(ph);
	cd_close();
	exit(0);
}

void bootrom(){
	float titi;
	struct timeval tvi,tvf;

	gettimeofday(&tvi,0);
	if (enable_dynarec) {
		do {dyn_handle_flow();} while (ctx.pc0!=0x80030000);
	} else {
		do {handle_flow();} while (ctx.pc0!=0x80030000);
	}
	gettimeofday(&tvf,0);
	titi=(tvf.tv_sec-tvi.tv_sec)+(tvf.tv_usec-tvi.tv_usec)/1000000.0;
	sprintf(distring,"Booted in %.3f seconds",titi);
	gui->print(distring);
	enable_vsync=1;
	ctx.nstp=0;
}

void text_fonc(char *filename,char *about) {
	printf("        %s: %s\n",filename,about);
}

int main(int argc,char *argv[]) {
	int i=1;
	int no_mop=0;

	signal(SIGINT,quit);

	if (strstr(argv[0],"gsope")==NULL)
		gui=&console_ui;
	else
		gui=&gtk_ui;

	initall();
	if (i!=argc) do {
		if (strncmp(argv[i],"-",1)==0) {
			if (strncmp(argv[i],"-i",2)==0) {
				i++;sscanf(argv[i],"%x",&halt_address);
			}
			if (strncmp(argv[i],"-p",2)==0) force_pal=1;
			if (strncmp(argv[i],"-s",2)==0) disable_sound=0;
			if (strncmp(argv[i],"-d",2)==0) {
				enable_dynarec=0;
			}
			if (strncmp(argv[i],"-h",2)==0) {
				print_help(argc,argv);
				return(0);
			}
			if (strncmp(argv[i],"-u",2)==0) norender=1;
			if (strncmp(argv[i],"-n",2)==0) notextures=1;
			if (strncmp(argv[i],"-b",2)==0) anti_alias=1;
			if (strncmp(argv[i],"-l",2)==0) {
				printf("Available plugins\n");
				printf("    Rendering:\n");
				list_renderer_plugins(text_fonc);
				printf("    Sound:\n");
				list_sound_plugins(text_fonc);
				return(0);
			}
			if (strncmp(argv[i],"-v=",3)==0) {
				sscanf(argv[i],"-v=%d",&ncyc);
				if (ncyc>65) ncyc=65;
				ncyc*=1000;
			}
        		if (strncmp(argv[i],"-m=",3)==0)
            		    sscanf(argv[i],"-m=%x",&debug_mask);
        		if (strncmp(argv[i],"-B",2)==0) {
			    boot_cdrom = 1;
			}
			i++;
		} else {
			no_mop=1;
		}
	} while ((i!=argc) && (no_mop==0));
	if (enable_dynarec) reset_dynarec();
	make_debug_flags();
	i--;
	gui->start_ui(argc-i,argv+i);
	return(-1);
}

void quit(int signal)
{
if (screen_mode!=0xff) rnd->exit();
if (disable_sound==0)  snd->exit();
cd_close();

printf("# Sope # shutting down by signal %d\n",signal);

exit(0);
}