#include <string.h>
#include <pthread.h>
#include "sope.h"
#include "hwr.h"
#include "sound.h"

int vag_depack(uint16 * entree, sint16 * sortie);

int unsigned voices;

typedef struct _Voice_ {
	sint16 volL, volR;
	uint16 freq;
	uint16 saddr;
	uint16 raddr;
	int on;
	int pos;
	int size;
	sint16 buffer[65536];
	uint32 count;
} Voice;

Voice chn[24];

void sp0write(short unsigned af)
{
}


void spudma()
{
	int unsigned i, s;
	uint8 *su, *cu;
	uint32 dma1, dma2, dma3;

	dma1 = HRM32(spu_madr);
	dma2 = HRM32(spu_bcr);
	dma3 = HRM32(spu_chcr);

	switch (dma3) {
/*SPU2CPU*/
	case 0x01000200:
		s = (dma2 >> 16) * 64;
		if (debug_sound)
			printf("SPU 2 CPU: copy %u bytes from %8.8x to %8.8x\n", s,
				   HRM16(spu_sb_addr) * 8, dma1);
		cu = RAOM + (dma1 & 0x1fffff);
		su = SBF + HRM16(spu_sb_addr) * 8;
		for (i = 0; i < s; i++) {
			*cu = *su;
			cu++;
			su++;
		}
		break;
/*CPU2SPU*/
	case 0x01000201:
		s = (dma2 >> 16) * 64;
		if (debug_sound)
			printf("CPU 2 SPU: copy %u bytes from %8.8x to %8.8x\n", s,
				   dma1, HRM16(spu_sb_addr) * 8);
		cu = RAOM + (dma1 & 0x1fffff);
		su = SBF + HRM16(spu_sb_addr) * 8;
		for (i = 0; i < s; i++) {
			*su = *cu;
			cu++;
			su++;
		}
		break;
	default:
		if (debug_sound)
			printf("SPU DMA function %8.8x not impl...\n", dma3);
	}
	HRM32(spu_chcr) &= 0xfeffffff;
	HRM16(sp0) &= 0xffdf;
	if (debug_sound)
		printf("DMA transfer OK\n");
}

void vonof(uint16 af, int p)
{
	int i;
	uint32 c;
	int taille;

	if (disable_sound)
		return;

	switch (p) {
	case 1:
//      if (af!=0) system("sleep 1");
		c = af;
		voices |= c;
		break;
	case 2:
		c = af << 16;
		voices |= c;
		break;
	case 3:
		c = 0xffffffff - af;
		voices &= c;
		break;
	case 4:
		c = 0xffffffff - (af << 16);
		voices &= c;
		break;
	default:
		c = 0;
		break;
	}
	if (p < 3) {
		if (debug_sound) {
			printf("# Sope # Voices:");
			for (i = 0; i < 24; i++) {
				if ((c >> i) & 1)
					printf(" O");
				else
					printf(" -");
			}
			printf("\n");
		}
		for (i = 0; i < 24; i++) {
			if ((c >> i) & 1) {
				taille =
					vag_depack((uint16 *) (SBF + chn[i].saddr),
							   chn[i].buffer);
				chn[i].on = 1;
				chn[i].size = taille;
				chn[i].pos = 0;
				chn[i].count = 0;
				//set_frequency(chn[i].freq);
				//play_sample(caca,taille);
				return;
			}
		}
	}

}

void setvoiceparam(uint16 hrn, uint16 af)
{
	int unsigned vn, vp;

	vn = (hrn - 0x1c00) >> 4;
	vp = (hrn - 0x1c00) & 15;
//if (debug_sound) printf("Voice param changed at no %u,%u,%u\n",vn,vp/2,af);

	switch (vp) {
	case 0x0:
		chn[vn].volL = af;
		return;
	case 0x2:
		chn[vn].volR = af;
		return;
	case 0x4:
		chn[vn].freq = af * 44100 / 4096;
		return;
	case 0x6:
		chn[vn].saddr = af * 8;
		return;
	case 0x8:
		return;
	case 0xa:
		return;
	case 0xc:
		return;
	case 0xe:
		chn[vn].raddr = af * 8;
		return;
	}

}

void mix_sound();

static void *sound_thread(void *arg)
{
	while (69) {
		mix_sound();
	}
}

static pthread_t thread;

void init_sound()
{

	if (disable_sound)
		return;
	snd->init();
	pthread_create(&thread, NULL, sound_thread, NULL);
	pthread_detach(thread);
}


#define SIZE_SBUF 1024
sint16 sbuf2[SIZE_SBUF];
sint32 sbuf[SIZE_SBUF];

void mix_sound()
{
	int i, j, n;
	Voice *ch;

	n = 0;
	for (i = 0; i < 24; i++) {
		ch = &chn[i];
		if (!ch->on)
			continue;
		n++;
		for (j = 0; j < SIZE_SBUF; j++) {
			ch->count += ch->freq;
			if (ch->count >= 44100) {
				ch->pos += (ch->count / 44100);
				ch->count %= 44100;
			}
			sbuf[j] += ch->buffer[ch->pos];
			if (ch->pos >= ch->size) {
				ch->on = 0;
				break;
			}
		}
	}

	if (n > 0) {
		for (i = 0; i < SIZE_SBUF; i++)
			sbuf2[i] = sbuf[i] / n;
	}
	snd->play_sample(sbuf2, SIZE_SBUF);
	memset(sbuf, 0, 4 * SIZE_SBUF);
	memset(sbuf2, 0, 2 * SIZE_SBUF);
}
